#define __USE_MINGW_ANSI_STDIO 1
#include <stdlib.h>
#include <intrin.h>
#include <malloc.h>
#include <windows.h>
#include <stdio.h>
#include <stdint.h>

 uint64_t min3_a (uint64_t rcx, uint64_t rdx, uint64_t r8);
 uint64_t min3_b (uint64_t rcx, uint64_t rdx, uint64_t r8);
 uint64_t min3_c (uint64_t rcx, uint64_t rdx, uint64_t r8);

//-----------------------------------------------------------------------------
//
//  queryPerformanceCounter - similar to QueryPerformanceCounter, but returns
//                            count directly.

uint64_t queryPerformanceCounter (void)
    {
    LARGE_INTEGER int64;
    QueryPerformanceCounter (&int64);
    return int64.QuadPart;
    }

//-----------------------------------------------------------------------------
//
// queryPerformanceFrequency - same as QueryPerformanceFrequency, but returns  count direcly.

uint64_t queryPerformanceFrequency (void)
    {
    LARGE_INTEGER int64;

    QueryPerformanceFrequency (&int64);
    return int64.QuadPart;
    }

//----------------------------------------------------------------------------
//
// lfsr64gpr - left shift galois type lfsr for 64-bit data, general purpose register implementation
//
static uint64_t lfsr64gpr (uint64_t data, uint64_t mask)
   {
   uint64_t carryOut = data >> 63;
   uint64_t maskOrZ = -carryOut; 
   return (data << 1) ^ (maskOrZ & mask);
   }

//---------------------------------------------------------------------------

static __m128 runtestu (float *buffer, uint64_t m128count, uint64_t loops)
    {
    uint64_t startCount, elapsed, index, loop;
    __m128 *buffer128 = (void *) buffer;
    __m128 total = _mm_set_ps (0, 0, 0, 0);
    double ns;

    startCount = queryPerformanceCounter ();
    for (loop = 0; loop < loops; loop++)
        for (index = 0; index < m128count; index++)
            {
            __m128 v0 = _mm_loadu_ps ((float *)&buffer128 [index + 0]);
            total = _mm_add_ps (total, v0);
            }

    elapsed = queryPerformanceCounter () - startCount;
    ns = (double) elapsed / queryPerformanceFrequency () * 1000000000 / loops / m128count;
    printf ("%7.2f ns\n", ns);
    return total;
    }

//---------------------------------------------------------------------------

static __m128 runtesta (float *buffer, uint64_t m128count, uint64_t loops)
    {
    uint64_t startCount, elapsed, index, loop;
    __m128 *buffer128 = (void *) buffer;
    __m128 total = _mm_set_ps (0, 0, 0, 0);
    double ns;

    startCount = queryPerformanceCounter ();
    for (loop = 0; loop < loops; loop++)
        for (index = 0; index < m128count; index++)
            {
            __m128 v0 = _mm_load_ps ((float *)&buffer128 [index + 0]);
            total = _mm_add_ps (total, v0);
            }

    elapsed = queryPerformanceCounter () - startCount;
    ns = (double) elapsed / queryPerformanceFrequency () * 1000000000 / loops / m128count;
    printf ("%7.2f ns\n", ns);
    return total;
    }

//---------------------------------------------------------------------------

int main (void)
    {
    uint64_t mask, pattern, index;
    uint64_t m128count = 1024, loops = 800000;
    float *buffer0, *buffer1;
    __m128 total = _mm_set_ps (0, 0, 0, 0);

    buffer0 = _aligned_malloc (sizeof total * m128count + 63, 64); 
    buffer1 = _aligned_malloc (sizeof total * m128count + 63, 64); 

    buffer0 = (void *) ((uint64_t) buffer0 + 0);
    buffer1 = (void *) ((uint64_t) buffer1 + 1);

    // fill the buffers with valid, normal floating point values
    mask = 0xBEFFFFFFFFFFFFFF;
    pattern = 1;
    for (index = 0; index < m128count * 4; index++)
        {
        float psrandom = (float) pattern;
        pattern = lfsr64gpr (pattern, mask);
        ((float *) buffer0) [index] = psrandom;
        ((float *) buffer1) [index] = psrandom;
        }

    // raise our priority to increase measurement accuracy
    SetPriorityClass (GetCurrentProcess (), REALTIME_PRIORITY_CLASS);

    total = _mm_add_ps (total, runtestu (buffer0, m128count, loops));
    total = _mm_add_ps (total, runtesta (buffer0, m128count, loops));
    total = _mm_add_ps (total, runtestu (buffer1, m128count, loops));

    // restore priority
    SetPriorityClass (GetCurrentProcess (), NORMAL_PRIORITY_CLASS);

    _aligned_free (buffer0);
    _aligned_free (buffer1);

    // to prevent optimizers from removing test function
    return _mm_cvtt_ss2si (total) == 99999;
    }

//---------------------------------------------------------------------------
